# https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#reserved_keywords',

RESERVED_KEYWORDS = [
    "ALL",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC",
    "ASSERT_ROWS_MODIFIED",
    "AT",
    "BETWEEN",
    "BY",
    "CASE",
    "CAST",
    "COLLATE",
    "CONTAINS",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "DEFAULT",
    "DEFINE",
    "DESC",
    "DISTINCT",
    "ELSE",
    "END",
    "ENUM",
    "ESCAPE",
    "EXCEPT",
    "EXCLUDE",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FOLLOWING",
    "FOR",
    "FROM",
    "FULL",
    "GROUP",
    "GROUPING",
    "GROUPS",
    "HASH",
    "HAVING",
    "IF",
    "IGNORE",
    "IN",
    "INNER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LOOKUP",
    "MERGE",
    "NATURAL",
    "NEW",
    "NO",
    "NOT",
    "NULL",
    "NULLS",
    "OF",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "OVER",
    "PARTITION",
    "PRECEDING",
    "PROTO",
    "QUALIFY",
    "RANGE",
    "RECURSIVE",
    "RESPECT",
    "RIGHT",
    "ROLLUP",
    "ROWS",
    "SELECT",
    "SET",
    "SOME",
    "STRUCT",
    "TABLESAMPLE",
    "THEN",
    "TO",
    "TREAT",
    "TRUE",
    "UNBOUNDED",
    "UNION",
    "UNNEST",
    "USING",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
    "WITHIN",
]



# Autogenerated using `scripts/parse_builtins.py`

BUILTIN_FUNCTIONS = [
    "ABS_BIGNUMERIC",
    "ABS_DOUBLE",
    "ABS_FLOAT",
    "ABS_INT",
    "ABS_NUMERIC",
    "ABS_UINT",
    "ACOSH_DOUBLE",
    "ACOS_DOUBLE",
    "ADD_BIGNUMERIC",
    "ADD_DATETIME_INTERVAL",
    "ADD_DATE_INT",
    "ADD_DATE_INTERVAL",
    "ADD_DOUBLE",
    "ADD_INT",
    "ADD_INTERVAL_DATE",
    "ADD_INTERVAL_DATETIME",
    "ADD_INTERVAL_INTERVAL",
    "ADD_INTERVAL_TIMESTAMP",
    "ADD_MONTHS_DATE",
    "ADD_MONTHS_DATETIME",
    "ADD_MONTHS_TIMESTAMP",
    "ADD_NUMERIC",
    "ADD_TIMESTAMP_INTERVAL",
    "ADD_UINT",
    "AEAD_DECRYPT_BYTES",
    "AEAD_DECRYPT_STRING",
    "AEAD_DECRYPT_STRUCT_BYTES",
    "AEAD_DECRYPT_STRUCT_STRING",
    "AEAD_ENCRYPT_BYTES",
    "AEAD_ENCRYPT_STRING",
    "AEAD_ENCRYPT_STRUCT_BYTES",
    "AEAD_ENCRYPT_STRUCT_STRING",
    "AEAD_ENVELOPE_DECRYPT_BYTES",
    "AEAD_ENVELOPE_DECRYPT_STRING",
    "AEAD_ENVELOPE_ENCRYPT_BYTES",
    "AEAD_ENVELOPE_ENCRYPT_STRING",
    "AND",
    "ANON_AVG_DOUBLE",
    "ANON_AVG_DOUBLE_WITH_REPORT_JSON",
    "ANON_AVG_DOUBLE_WITH_REPORT_PROTO",
    "ANON_AVG_NUMERIC",
    "ANON_COUNT",
    "ANON_COUNT_STAR",
    "ANON_COUNT_STAR_WITH_REPORT_JSON",
    "ANON_COUNT_STAR_WITH_REPORT_PROTO",
    "ANON_COUNT_WITH_REPORT_JSON",
    "ANON_COUNT_WITH_REPORT_PROTO",
    "ANON_PERCENTILE_CONT_DOUBLE",
    "ANON_PERCENTILE_CONT_DOUBLE_ARRAY",
    "ANON_QUANTILES_DOUBLE",
    "ANON_QUANTILES_DOUBLE_ARRAY",
    "ANON_QUANTILES_DOUBLE_ARRAY_WITH_REPORT_JSON",
    "ANON_QUANTILES_DOUBLE_ARRAY_WITH_REPORT_PROTO",
    "ANON_QUANTILES_DOUBLE_WITH_REPORT_JSON",
    "ANON_QUANTILES_DOUBLE_WITH_REPORT_PROTO",
    "ANON_STDDEV_POP_DOUBLE",
    "ANON_STDDEV_POP_DOUBLE_ARRAY",
    "ANON_SUM_DOUBLE",
    "ANON_SUM_INT",
    "ANON_SUM_NUMERIC",
    "ANON_SUM_UINT",
    "ANON_SUM_WITH_REPORT_JSON_DOUBLE",
    "ANON_SUM_WITH_REPORT_JSON_INT",
    "ANON_SUM_WITH_REPORT_JSON_UINT",
    "ANON_SUM_WITH_REPORT_PROTO_DOUBLE",
    "ANON_SUM_WITH_REPORT_PROTO_INT",
    "ANON_SUM_WITH_REPORT_PROTO_UINT",
    "ANON_VAR_POP_DOUBLE",
    "ANON_VAR_POP_DOUBLE_ARRAY",
    "ANY_VALUE",
    "APPROXIMATE_JACCARD_INDEX",
    "APPROXIMATE_SIMILARITY",
    "APPROX_COUNT_DISTINCT",
    "APPROX_PERCENTILE",
    "APPROX_PERCENTILE_ACCUMULATE",
    "APPROX_PERCENTILE_COMBINE",
    "APPROX_PERCENTILE_ESTIMATE",
    "APPROX_QUANTILES",
    "APPROX_TOP_COUNT",
    "APPROX_TOP_K",
    "APPROX_TOP_K_ACCUMULATE",
    "APPROX_TOP_K_COMBINE",
    "APPROX_TOP_K_ESTIMATE",
    "APPROX_TOP_SUM_BIGNUMERIC",
    "APPROX_TOP_SUM_DOUBLE",
    "APPROX_TOP_SUM_INT",
    "APPROX_TOP_SUM_NUMERIC",
    "APPROX_TOP_SUM_UINT",
    "ARRAYS_OVERLAP",
    "ARRAY_AGG",
    "ARRAY_APPEND",
    "ARRAY_AT_OFFSET",
    "ARRAY_AT_ORDINAL",
    "ARRAY_AVG_BIGNUMERIC",
    "ARRAY_AVG_DOUBLE",
    "ARRAY_AVG_FLOAT",
    "ARRAY_AVG_INT",
    "ARRAY_AVG_INTERVAL",
    "ARRAY_AVG_NUMERIC",
    "ARRAY_AVG_UINT",
    "ARRAY_CAT",
    "ARRAY_COMPACT",
    "ARRAY_CONCAT",
    "ARRAY_CONCAT_AGG",
    "ARRAY_CONCAT_OP",
    "ARRAY_CONSTRUCT",
    "ARRAY_CONSTRUCT_COMPACT",
    "ARRAY_CONTAINS",
    "ARRAY_FILTER",
    "ARRAY_FILTER_WITH_INDEX",
    "ARRAY_FIND",
    "ARRAY_FIND_ALL",
    "ARRAY_FIND_ALL_LAMBDA",
    "ARRAY_FIND_LAMBDA",
    "ARRAY_FIRST",
    "ARRAY_FIRST_N",
    "ARRAY_INCLUDES",
    "ARRAY_INCLUDES_ALL",
    "ARRAY_INCLUDES_ANY",
    "ARRAY_INCLUDES_LAMBDA",
    "ARRAY_INSERT",
    "ARRAY_INTERSECTION",
    "ARRAY_IS_DISTINCT",
    "ARRAY_LAST",
    "ARRAY_LAST_N",
    "ARRAY_LENGTH",
    "ARRAY_MAX",
    "ARRAY_MAX_DOUBLE",
    "ARRAY_MAX_FLOAT",
    "ARRAY_MIN",
    "ARRAY_OFFSET",
    "ARRAY_OFFSETS",
    "ARRAY_OFFSETS_LAMBDA",
    "ARRAY_OFFSET_LAMBDA",
    "ARRAY_POSITION",
    "ARRAY_PREPEND",
    "ARRAY_REMOVE_FIRST_N",
    "ARRAY_REMOVE_LAST_N",
    "ARRAY_REVERSE",
    "ARRAY_SIZE_ARRAY",
    "ARRAY_SIZE_VARIANT",
    "ARRAY_SLICE",
    "ARRAY_SUM_BIGNUMERIC",
    "ARRAY_SUM_DOUBLE",
    "ARRAY_SUM_FLOAT",
    "ARRAY_SUM_INT",
    "ARRAY_SUM_INTERVAL",
    "ARRAY_SUM_NUMERIC",
    "ARRAY_SUM_UINT",
    "ARRAY_TO_BYTES",
    "ARRAY_TO_STRING",
    "ARRAY_TRANSFORM",
    "ARRAY_TRANSFORM_WITH_INDEX",
    "ASCII_BYTES",
    "ASCII_STRING",
    "ASINH_DOUBLE",
    "ASIN_DOUBLE",
    "AS_ARRAY",
    "AS_BOOLEAN",
    "AS_CHAR",
    "AS_DATE",
    "AS_DECIMAL",
    "AS_DOUBLE",
    "AS_INTEGER",
    "AS_OBJECT",
    "AS_TIME",
    "AS_TIMESTAMP",
    "ATAN",
    "ATANH_DOUBLE",
    "ATAN_DOUBLE",
    "AVG_BIGNUMERIC",
    "AVG_DOUBLE",
    "AVG_INT",
    "AVG_INTERVAL",
    "AVG_NUMERIC",
    "AVG_UINT",
    "BASE",
    "BETWEEN",
    "BETWEEN_INT",
    "BETWEEN_UINT",
    "BITAND_AGG",
    "BITOR_AGG",
    "BITWISE_AND_BYTES",
    "BITWISE_AND_INT",
    "BITWISE_AND_UINT",
    "BITWISE_LEFT_SHIFT_BYTES",
    "BITWISE_LEFT_SHIFT_INT",
    "BITWISE_LEFT_SHIFT_UINT",
    "BITWISE_NOT_BYTES",
    "BITWISE_NOT_INT",
    "BITWISE_NOT_UINT",
    "BITWISE_OR_BYTES",
    "BITWISE_OR_INT",
    "BITWISE_OR_UINT",
    "BITWISE_RIGHT_SHIFT_BYTES",
    "BITWISE_RIGHT_SHIFT_INT",
    "BITWISE_RIGHT_SHIFT_UINT",
    "BITWISE_XOR_BYTES",
    "BITWISE_XOR_INT",
    "BITWISE_XOR_UINT",
    "BITXOR_AGG",
    "BIT_AND_DIFF_ARGS",
    "BIT_AND_INT",
    "BIT_AND_SAME_ARGS",
    "BIT_AND_UINT",
    "BIT_BIT_SHIFT_RIGHT",
    "BIT_CAST_INT",
    "BIT_CAST_UINT",
    "BIT_COUNT_BYTES",
    "BIT_COUNT_INT",
    "BIT_COUNT_UINT",
    "BIT_NOT",
    "BIT_OR_DIFF_ARGS",
    "BIT_OR_INT",
    "BIT_OR_SAME_ARGS",
    "BIT_OR_UINT",
    "BIT_SHIFT_LEFT",
    "BIT_XOR_DIFF_ARGS",
    "BIT_XOR_INT",
    "BIT_XOR_SAME_ARGS",
    "BIT_XOR_UINT",
    "BOOLAND_AGG",
    "BOOLAND_DIFF_ARGS",
    "BOOLAND_SAME_ARGS",
    "BOOLNOT",
    "BOOLOR_AGG",
    "BOOLOR_DIFF_ARGS",
    "BOOLOR_SAME_ARGS",
    "BOOLXOR_AGG",
    "BOOLXOR_DIFF_ARGS",
    "BOOLXOR_SAME_ARGS",
    "BYTE_ARRAY_LIKE_ALL",
    "BYTE_ARRAY_LIKE_ANY",
    "BYTE_LENGTH_BYTES",
    "BYTE_LENGTH_STRING",
    "BYTE_LIKE",
    "BYTE_LIKE_ALL",
    "BYTE_LIKE_ANY",
    "CASE_NO_VALUE",
    "CASE_WITH_VALUE",
    "CBRT_BIGNUMERIC",
    "CBRT_DOUBLE",
    "CBRT_NUMERIC",
    "CEIL_BIGNUMERIC",
    "CEIL_DOUBLE",
    "CEIL_FLOAT",
    "CEIL_NUMERIC",
    "CHAR_LENGTH_STRING",
    "CHECK_JSON_STRING",
    "CHECK_JSON_VARIANT",
    "CHECK_XML",
    "CHR_STRING",
    "COALESCE",
    "CODE_POINTS_TO_BYTES",
    "CODE_POINTS_TO_STRING",
    "COLLATE",
    "CONCAT_BYTES",
    "CONCAT_OP_BYTES",
    "CONCAT_OP_STRING",
    "CONCAT_STRING",
    "CONTAINS",
    "CONTAINS_KEY",
    "CONVERT_TIMESTAMP",
    "CONVERT_TIMESTAMP_NTZ",
    "CORR",
    "CORR_BIGNUMERIC",
    "CORR_NUMERIC",
    "COSH_DOUBLE",
    "COS_DOUBLE",
    "COTH_DOUBLE",
    "COT_DOUBLE",
    "COUNT",
    "COUNTIF",
    "COUNT_STAR",
    "COVAR_POP",
    "COVAR_POP_BIGNUMERIC",
    "COVAR_POP_NUMERIC",
    "COVAR_SAMP",
    "COVAR_SAMP_BIGNUMERIC",
    "COVAR_SAMP_NUMERIC",
    "CSCH_DOUBLE",
    "CSC_DOUBLE",
    "CUME_DIST",
    "CURRENT_DATE",
    "CURRENT_DATETIME",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "D",
    "DATEADD_DATE",
    "DATEADD_TIME",
    "DATEADD_TIMESTAMP",
    "DATEDIFF_DATE",
    "DATEDIFF_TIMESTAMP",
    "DATETIME_ADD",
    "DATETIME_BUCKET",
    "DATETIME_DIFF",
    "DATETIME_FROM_DATE",
    "DATETIME_FROM_DATETIME",
    "DATETIME_FROM_DATE_AND_TIME",
    "DATETIME_FROM_STRING",
    "DATETIME_FROM_TIMESTAMP",
    "DATETIME_FROM_YEAR_MONTH_DAY_HOUR_MINUTE_SECOND",
    "DATETIME_SUB",
    "DATETIME_TRUNC",
    "DATE_ADD_DATE",
    "DATE_BUCKET",
    "DATE_DIFF_DATE",
    "DATE_FROM_DATE",
    "DATE_FROM_DATETIME",
    "DATE_FROM_PARTS",
    "DATE_FROM_STRING",
    "DATE_FROM_TIMESTAMP",
    "DATE_FROM_UNIX_DATE",
    "DATE_FROM_YEAR_MONTH_DAY",
    "DATE_PART_DATE",
    "DATE_PART_TIME",
    "DATE_PART_TIMESTAMP",
    "DATE_SUB_DATE",
    "DATE_TRUNC_DATE",
    "DATE_TRUNC_TIME",
    "DATE_TRUNC_TIMESTAMP",
    "DAYNAME_DATE",
    "DAYNAME_DATETIME",
    "DAYNAME_TIMESTAMP",
    "DECIMAL_LOGARITHM_BIGNUMERIC",
    "DECIMAL_LOGARITHM_DOUBLE",
    "DECIMAL_LOGARITHM_NUMERIC",
    "DECODE",
    "DEGREES",
    "DENSE_RANK",
    "DETERMINISTIC_DECRYPT_BYTES",
    "DETERMINISTIC_DECRYPT_STRING",
    "DETERMINISTIC_DECRYPT_STRUCT_BYTES",
    "DETERMINISTIC_DECRYPT_STRUCT_STRING",
    "DETERMINISTIC_ENCRYPT_BYTES",
    "DETERMINISTIC_ENCRYPT_STRING",
    "DETERMINISTIC_ENCRYPT_STRUCT_BYTES",
    "DETERMINISTIC_ENCRYPT_STRUCT_STRING",
    "DIFFERENTIAL_PRIVACY_AVG_DOUBLE",
    "DIFFERENTIAL_PRIVACY_AVG_DOUBLE_REPORT_JSON",
    "DIFFERENTIAL_PRIVACY_AVG_DOUBLE_REPORT_PROTO",
    "DIFFERENTIAL_PRIVACY_AVG_NUMERIC",
    "DIFFERENTIAL_PRIVACY_COUNT",
    "DIFFERENTIAL_PRIVACY_COUNT_REPORT_JSON",
    "DIFFERENTIAL_PRIVACY_COUNT_REPORT_PROTO",
    "DIFFERENTIAL_PRIVACY_COUNT_STAR",
    "DIFFERENTIAL_PRIVACY_COUNT_STAR_REPORT_JSON",
    "DIFFERENTIAL_PRIVACY_COUNT_STAR_REPORT_PROTO",
    "DIFFERENTIAL_PRIVACY_PERCENTILE_CONT_DOUBLE",
    "DIFFERENTIAL_PRIVACY_PERCENTILE_CONT_DOUBLE_ARRAY",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE_ARRAY",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE_ARRAY_REPORT_JSON",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE_ARRAY_REPORT_PROTO",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE_REPORT_JSON",
    "DIFFERENTIAL_PRIVACY_QUANTILES_DOUBLE_REPORT_PROTO",
    "DIFFERENTIAL_PRIVACY_STDDEV_POP_DOUBLE",
    "DIFFERENTIAL_PRIVACY_STDDEV_POP_DOUBLE_ARRAY",
    "DIFFERENTIAL_PRIVACY_SUM_DOUBLE",
    "DIFFERENTIAL_PRIVACY_SUM_INT",
    "DIFFERENTIAL_PRIVACY_SUM_NUMERIC",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_JSON_DOUBLE",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_JSON_INT",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_JSON_UINT",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_PROTO_DOUBLE",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_PROTO_INT",
    "DIFFERENTIAL_PRIVACY_SUM_REPORT_PROTO_UINT",
    "DIFFERENTIAL_PRIVACY_SUM_UINT",
    "DIFFERENTIAL_PRIVACY_VAR_POP_DOUBLE",
    "DIFFERENTIAL_PRIVACY_VAR_POP_DOUBLE_ARRAY",
    "DISTINCT",
    "DISTINCT_INT",
    "DISTINCT_UINT",
    "DIVIDE_BIGNUMERIC",
    "DIVIDE_DOUBLE",
    "DIVIDE_INTERVAL_INT",
    "DIVIDE_NUMERIC",
    "DIV_BIGNUMERIC",
    "DIV_INT",
    "DIV_NUMERIC",
    "DIV_UINT",
    "ENDSWITH",
    "ENDS_WITH_BYTES",
    "ENDS_WITH_STRING",
    "ENUM_VALUE_DESCRIPTOR_PROTO",
    "EQUAL",
    "EQUAL_BOOL_STRING",
    "EQUAL_DOUBLE_STRING",
    "EQUAL_FLOAT_STRING",
    "EQUAL_INT",
    "EQUAL_NUMERIC_STRING",
    "EQUAL_STRING_BOOL",
    "EQUAL_STRING_DOUBLE",
    "EQUAL_STRING_FLOAT",
    "EQUAL_STRING_INT",
    "EQUAL_STRING_NUMERIC",
    "EQUAL_UINT",
    "ERROR",
    "EXP_BIGNUMERIC",
    "EXP_DOUBLE",
    "EXP_NUMERIC",
    "EXTRACT_DATETIME_FROM_TIMESTAMP",
    "EXTRACT_DATE_FROM_DATETIME",
    "EXTRACT_DATE_FROM_TIMESTAMP",
    "EXTRACT_FROM_DATE",
    "EXTRACT_FROM_DATETIME",
    "EXTRACT_FROM_INTERVAL",
    "EXTRACT_FROM_TIME",
    "EXTRACT_FROM_TIMESTAMP",
    "EXTRACT_PART",
    "EXTRACT_TIME_FROM_DATETIME",
    "EXTRACT_TIME_FROM_TIMESTAMP",
    "FACTORIAL",
    "FARM_FINGERPRINT_BYTES",
    "FARM_FINGERPRINT_STRING",
    "FIRST_VALUE",
    "FLATTEN",
    "FLOOR_BIGNUMERIC",
    "FLOOR_DOUBLE",
    "FLOOR_FLOAT",
    "FLOOR_NUMERIC",
    "FORMAT_DATE",
    "FORMAT_DATETIME",
    "FORMAT_STRING",
    "FORMAT_TIME",
    "FORMAT_TIMESTAMP",
    "FROM_BASE",
    "FROM_HEX",
    "FROM_PROTO_BOOL",
    "FROM_PROTO_BYTES",
    "FROM_PROTO_DATE",
    "FROM_PROTO_DOUBLE",
    "FROM_PROTO_FLOAT",
    "FROM_PROTO_IDEMPOTENT_BOOL",
    "FROM_PROTO_IDEMPOTENT_BYTES",
    "FROM_PROTO_IDEMPOTENT_DATE",
    "FROM_PROTO_IDEMPOTENT_DOUBLE",
    "FROM_PROTO_IDEMPOTENT_FLOAT",
    "FROM_PROTO_IDEMPOTENT_INT",
    "FROM_PROTO_IDEMPOTENT_STRING",
    "FROM_PROTO_IDEMPOTENT_TIME",
    "FROM_PROTO_IDEMPOTENT_TIMESTAMP",
    "FROM_PROTO_IDEMPOTENT_UINT",
    "FROM_PROTO_INT",
    "FROM_PROTO_STRING",
    "FROM_PROTO_TIMESTAMP",
    "FROM_PROTO_TIME_OF_DAY",
    "FROM_PROTO_UINT",
    "GENERATE_ARRAY_BIGNUMERIC",
    "GENERATE_ARRAY_DOUBLE",
    "GENERATE_ARRAY_INT",
    "GENERATE_ARRAY_NUMERIC",
    "GENERATE_ARRAY_UINT",
    "GENERATE_DATETIME_RANGE_ARRAY",
    "GENERATE_DATE_ARRAY",
    "GENERATE_DATE_RANGE_ARRAY",
    "GENERATE_TIMESTAMP_ARRAY",
    "GENERATE_TIMESTAMP_RANGE_ARRAY",
    "GENERATE_UUID",
    "GET_ARRAY",
    "GET_OBJECT",
    "GET_OBJECT_STRING",
    "GET_PATH",
    "GET_PATH_OP_VARIANT",
    "GET_VARIANT_INT",
    "GREATER",
    "GREATER_INT",
    "GREATER_OR_EQUAL",
    "GREATER_OR_EQUAL_INT",
    "GREATER_OR_EQUAL_UINT",
    "GREATER_UINT",
    "GREATEST",
    "GROUPING",
    "GROUPING_ID",
    "HASH_AGG",
    "HAVERSINE",
    "HLL",
    "HLL_ACCUMULATE",
    "HLL_COMBINE",
    "HLL_COUNT_EXTRACT",
    "HLL_COUNT_INIT_BIGNUMERIC",
    "HLL_COUNT_INIT_BYTES",
    "HLL_COUNT_INIT_INT",
    "HLL_COUNT_INIT_NUMERIC",
    "HLL_COUNT_INIT_STRING",
    "HLL_COUNT_INIT_UINT",
    "HLL_COUNT_MERGE",
    "HLL_COUNT_MERGE_PARTIAL",
    "HLL_ESTIMATE",
    "HLL_EXPORT",
    "HLL_IMPORT",
    "HOUR_TIME",
    "HOUR_TIMESTAMP",
    "IEEE_DIVIDE_DOUBLE",
    "IEEE_DIVIDE_FLOAT",
    "IF",
    "IFERROR",
    "IFF",
    "IFNULL",
    "IN",
    "INITCAP_STRING",
    "INSERT",
    "INSTR_BYTES",
    "INSTR_STRING",
    "INTERVAL_CONSTRUCTOR",
    "INVALID_FUNCTION_ID",
    "IN_ARRAY",
    "ISERROR",
    "IS_ARRAY",
    "IS_BINARY",
    "IS_BOOLEAN",
    "IS_CHAR",
    "IS_DATE",
    "IS_DECIMAL",
    "IS_DOUBLE",
    "IS_FALSE",
    "IS_INF",
    "IS_INTEGER",
    "IS_NAN",
    "IS_NULL",
    "IS_NULL_VALUE",
    "IS_OBJECT",
    "IS_TIME",
    "IS_TIMESTAMP",
    "IS_TRUE",
    "JSON_ARRAY",
    "JSON_ARRAY_APPEND",
    "JSON_ARRAY_INSERT",
    "JSON_EXTRACT",
    "JSON_EXTRACT_ARRAY",
    "JSON_EXTRACT_ARRAY_JSON",
    "JSON_EXTRACT_JSON",
    "JSON_EXTRACT_SCALAR",
    "JSON_EXTRACT_SCALAR_JSON",
    "JSON_EXTRACT_STRING_ARRAY",
    "JSON_EXTRACT_STRING_ARRAY_JSON",
    "JSON_LAX_TO_BOOL",
    "JSON_LAX_TO_DOUBLE",
    "JSON_LAX_TO_INT",
    "JSON_LAX_TO_STRING",
    "JSON_OBJECT",
    "JSON_OBJECT_ARRAYS",
    "JSON_QUERY",
    "JSON_QUERY_ARRAY",
    "JSON_QUERY_ARRAY_JSON",
    "JSON_QUERY_JSON",
    "JSON_REMOVE",
    "JSON_SET",
    "JSON_STRIP_NULLS",
    "JSON_SUBSCRIPT_INT",
    "JSON_SUBSCRIPT_STRING",
    "JSON_TO_BOOL",
    "JSON_TO_DOUBLE",
    "JSON_TO_INT",
    "JSON_TO_STRING",
    "JSON_TYPE",
    "JSON_VALUE",
    "JSON_VALUE_ARRAY",
    "JSON_VALUE_ARRAY_JSON",
    "JSON_VALUE_JSON",
    "JUSTIFY_DAYS",
    "JUSTIFY_HOURS",
    "JUSTIFY_INTERVAL",
    "KEYS_ADD_KEY_FROM_RAW_BYTES",
    "KEYS_KEYSET_CHAIN_STRING_BYTES",
    "KEYS_KEYSET_CHAIN_STRING_BYTES_BYTES",
    "KEYS_KEYSET_FROM_JSON",
    "KEYS_KEYSET_LENGTH",
    "KEYS_KEYSET_TO_JSON",
    "KEYS_NEW_KEYSET",
    "KEYS_NEW_WRAPPED_KEYSET",
    "KEYS_REWRAP_KEYSET",
    "KEYS_ROTATE_KEYSET",
    "KEYS_ROTATE_WRAPPED_KEYSET",
    "KLL_QUANTILES_EXTRACT_DOUBLE",
    "KLL_QUANTILES_EXTRACT_INT",
    "KLL_QUANTILES_EXTRACT_POINT_DOUBLE",
    "KLL_QUANTILES_EXTRACT_POINT_INT",
    "KLL_QUANTILES_EXTRACT_POINT_UINT",
    "KLL_QUANTILES_EXTRACT_UINT",
    "KLL_QUANTILES_INIT_DOUBLE",
    "KLL_QUANTILES_INIT_INT",
    "KLL_QUANTILES_INIT_UINT",
    "KLL_QUANTILES_MERGE_DOUBLE",
    "KLL_QUANTILES_MERGE_INT",
    "KLL_QUANTILES_MERGE_PARTIAL",
    "KLL_QUANTILES_MERGE_POINT_DOUBLE",
    "KLL_QUANTILES_MERGE_POINT_INT",
    "KLL_QUANTILES_MERGE_POINT_UINT",
    "KLL_QUANTILES_MERGE_UINT",
    "KMS_DECRYPT_BYTES",
    "KMS_DECRYPT_STRING",
    "KMS_ENCRYPT_BYTES",
    "KMS_ENCRYPT_STRING",
    "KURTOSIS",
    "LAG",
    "LAST_DAY_DATE",
    "LAST_DAY_DATETIME",
    "LAST_DAY_TIMESTAMP",
    "LAST_VALUE",
    "LEAD",
    "LEAST",
    "LEFT_BYTES",
    "LEFT_STRING",
    "LENGTH_BYTES",
    "LENGTH_STRING",
    "LESS",
    "LESS_INT",
    "LESS_OR_EQUAL",
    "LESS_OR_EQUAL_INT",
    "LESS_OR_EQUAL_UINT",
    "LESS_UINT",
    "LISTAGG",
    "LOGARITHM_BIGNUMERIC",
    "LOGARITHM_DOUBLE",
    "LOGARITHM_NUMERIC",
    "LOGICAL_AND",
    "LOGICAL_OR",
    "LOWER_BYTES",
    "LOWER_STRING",
    "LPAD_BYTES",
    "LPAD_STRING",
    "LTRIM_BYTES",
    "LTRIM_STRING",
    "MAKE_ARRAY",
    "MAKE_INTERVAL",
    "MAX",
    "MD",
    "MEDIAN",
    "MIN",
    "MINHASH",
    "MINHASH_COMBINE",
    "MINUTE_TIME",
    "MINUTE_TIMESTAMP",
    "MODE",
    "MODIFY_MAP",
    "MOD_BIGNUMERIC",
    "MOD_INT",
    "MOD_NUMERIC",
    "MOD_OP",
    "MOD_UINT",
    "MONTHNAME_DATE",
    "MONTHNAME_DATETIME",
    "MONTHNAME_TIMESTAMP",
    "MULTIPLY_BIGNUMERIC",
    "MULTIPLY_DOUBLE",
    "MULTIPLY_INT",
    "MULTIPLY_INTERVAL_INT",
    "MULTIPLY_NUMERIC",
    "MULTIPLY_UINT",
    "NATURAL_LOGARITHM_BIGNUMERIC",
    "NATURAL_LOGARITHM_DOUBLE",
    "NATURAL_LOGARITHM_NUMERIC",
    "NET_FORMAT_IP",
    "NET_FORMAT_PACKED_IP",
    "NET_HOST",
    "NET_IPV",
    "NET_IP_FROM_STRING",
    "NET_IP_IN_NET",
    "NET_IP_NET_MASK",
    "NET_IP_TO_STRING",
    "NET_IP_TRUNC",
    "NET_MAKE_NET",
    "NET_PARSE_IP",
    "NET_PARSE_PACKED_IP",
    "NET_PUBLIC_SUFFIX",
    "NET_REG_DOMAIN",
    "NET_SAFE_IP_FROM_STRING",
    "NEXT_DAY_DATE",
    "NEXT_DAY_DATETIME",
    "NEXT_DAY_TIMESTAMP",
    "NORMALIZE_AND_CASEFOLD_STRING",
    "NORMALIZE_STRING",
    "NOT",
    "NOT_DISTINCT",
    "NOT_DISTINCT_INT",
    "NOT_DISTINCT_UINT",
    "NOT_EQUAL",
    "NOT_EQUAL_INT",
    "NOT_EQUAL_UINT",
    "NTH_VALUE",
    "NTILE",
    "NULLIF",
    "NULLIFERROR",
    "NULLIFZERO_BIGNUMERIC",
    "NULLIFZERO_DOUBLE",
    "NULLIFZERO_FLOAT",
    "NULLIFZERO_INT",
    "NULLIFZERO_NUMERIC",
    "NULLIFZERO_UINT",
    "NVL",
    "OBJECT_AGG",
    "OBJECT_CONSTRUCT",
    "OBJECT_DELETE",
    "OBJECT_INSERT",
    "OR",
    "PARSE_BIGNUMERIC",
    "PARSE_DATE",
    "PARSE_DATETIME",
    "PARSE_JSON",
    "PARSE_JSON_SNOWFLAKE",
    "PARSE_NUMERIC",
    "PARSE_TIME",
    "PARSE_TIMESTAMP",
    "PARSE_XML_STRING",
    "PARSE_XML_VARIANT",
    "PERCENTILE_CONT",
    "PERCENTILE_CONT_BIGNUMERIC",
    "PERCENTILE_CONT_NUMERIC",
    "PERCENTILE_DISC",
    "PERCENTILE_DISC_BIGNUMERIC",
    "PERCENTILE_DISC_NUMERIC",
    "PERCENT_RANK",
    "PI",
    "PI_BIGNUMERIC",
    "PI_DOUBLE",
    "PI_NUMERIC",
    "POW_BIGNUMERIC",
    "POW_DOUBLE",
    "POW_NUMERIC",
    "PREVIOUS_DAY_DATE",
    "PREVIOUS_DAY_DATETIME",
    "PREVIOUS_DAY_TIMESTAMP",
    "PROTO_DEFAULT_IF_NULL",
    "PROTO_MAP_AT_KEY",
    "RADIANS",
    "RAND",
    "RANDOM",
    "RANDSTR_DIFF_ARGS",
    "RANDSTR_SAME_ARGS",
    "RANGE",
    "RANGE_BUCKET",
    "RANGE_CONTAINS_ELEMENT",
    "RANGE_CONTAINS_RANGE",
    "RANGE_END",
    "RANGE_INTERSECT",
    "RANGE_IS_END_UNBOUNDED",
    "RANGE_IS_START_UNBOUNDED",
    "RANGE_OVERLAPS",
    "RANGE_START",
    "RANK",
    "REGEXP_CONTAINS_BYTES",
    "REGEXP_CONTAINS_STRING",
    "REGEXP_COUNT",
    "REGEXP_EXTRACT_ALL_BYTES",
    "REGEXP_EXTRACT_ALL_STRING",
    "REGEXP_EXTRACT_BYTES",
    "REGEXP_EXTRACT_STRING",
    "REGEXP_INSTR_BYTES",
    "REGEXP_INSTR_STRING",
    "REGEXP_LIKE",
    "REGEXP_MATCH_BYTES",
    "REGEXP_MATCH_STRING",
    "REGEXP_REPLACE_BYTES",
    "REGEXP_REPLACE_STRING",
    "REGEXP_SUBSTR_ALL",
    "REGR_AVGX_DIFF_ARGS",
    "REGR_AVGX_SAME_ARGS",
    "REGR_AVGY_DIFF_ARGS",
    "REGR_AVGY_SAME_ARGS",
    "REGR_COUNT_DIFF_ARGS",
    "REGR_COUNT_SAME_ARGS",
    "REGR_INTERCEPT",
    "REGR_R",
    "REGR_SLOPE",
    "REGR_SXX",
    "REGR_SYY",
    "REGR_VALX",
    "REGR_VALY",
    "REPEAT_BYTES",
    "REPEAT_STRING",
    "REPLACE_BYTES",
    "REPLACE_STRING",
    "REVERSE_BYTES",
    "REVERSE_STRING",
    "RIGHT_BYTES",
    "RIGHT_STRING",
    "ROUND_BIGNUMERIC",
    "ROUND_DOUBLE",
    "ROUND_FLOAT",
    "ROUND_NUMERIC",
    "ROUND_WITH_DIGITS_BIGNUMERIC",
    "ROUND_WITH_DIGITS_DOUBLE",
    "ROUND_WITH_DIGITS_FLOAT",
    "ROUND_WITH_DIGITS_NUMERIC",
    "ROUND_WITH_ROUNDING_MODE_BIGNUMERIC",
    "ROUND_WITH_ROUNDING_MODE_NUMERIC",
    "ROW_NUMBER",
    "RPAD_BYTES",
    "RPAD_STRING",
    "RTRIM_BYTES",
    "RTRIM_STRING",
    "S",
    "SAFE_ADD_BIGNUMERIC",
    "SAFE_ADD_DOUBLE",
    "SAFE_ADD_INT",
    "SAFE_ADD_NUMERIC",
    "SAFE_ADD_UINT",
    "SAFE_ARRAY_AT_OFFSET",
    "SAFE_ARRAY_AT_ORDINAL",
    "SAFE_CONVERT_BYTES_TO_STRING",
    "SAFE_DIVIDE_BIGNUMERIC",
    "SAFE_DIVIDE_DOUBLE",
    "SAFE_DIVIDE_NUMERIC",
    "SAFE_MULTIPLY_BIGNUMERIC",
    "SAFE_MULTIPLY_DOUBLE",
    "SAFE_MULTIPLY_INT",
    "SAFE_MULTIPLY_NUMERIC",
    "SAFE_MULTIPLY_UINT",
    "SAFE_PROTO_MAP_AT_KEY",
    "SAFE_SUBTRACT_BIGNUMERIC",
    "SAFE_SUBTRACT_DOUBLE",
    "SAFE_SUBTRACT_INT",
    "SAFE_SUBTRACT_NUMERIC",
    "SAFE_SUBTRACT_UINT",
    "SAFE_UNARY_MINUS_BIGNUMERIC",
    "SAFE_UNARY_MINUS_DOUBLE",
    "SAFE_UNARY_MINUS_FLOAT",
    "SAFE_UNARY_MINUS_INT",
    "SAFE_UNARY_MINUS_NUMERIC",
    "SECH_DOUBLE",
    "SECOND_TIME",
    "SECOND_TIMESTAMP",
    "SEC_DOUBLE",
    "SEQ",
    "SESSION_USER",
    "SHA",
    "SIGN_BIGNUMERIC",
    "SIGN_DOUBLE",
    "SIGN_FLOAT",
    "SIGN_INT",
    "SIGN_NUMERIC",
    "SIGN_UINT",
    "SINH_DOUBLE",
    "SIN_DOUBLE",
    "SKEW",
    "SOUNDEX_STRING",
    "SPLIT_BYTES",
    "SPLIT_STRING",
    "SQRT_BIGNUMERIC",
    "SQRT_DOUBLE",
    "SQRT_NUMERIC",
    "SQUARE",
    "STARTS_WITH_BYTES",
    "STARTS_WITH_STRING",
    "STDDEV_POP",
    "STDDEV_POP_BIGNUMERIC",
    "STDDEV_POP_NUMERIC",
    "STDDEV_SAMP",
    "STDDEV_SAMP_BIGNUMERIC",
    "STDDEV_SAMP_NUMERIC",
    "STRING_AGG_BYTES",
    "STRING_AGG_DELIM_BYTES",
    "STRING_AGG_DELIM_STRING",
    "STRING_AGG_STRING",
    "STRING_ARRAY_LIKE_ALL",
    "STRING_ARRAY_LIKE_ANY",
    "STRING_FROM_DATE",
    "STRING_FROM_DATETIME",
    "STRING_FROM_TIME",
    "STRING_FROM_TIMESTAMP",
    "STRING_LIKE",
    "STRING_LIKE_ALL",
    "STRING_LIKE_ANY",
    "STRIP_NULL_VALUE",
    "STRPOS_BYTES",
    "STRPOS_STRING",
    "STRTOK_TO_ARRAY",
    "ST_ACCUM",
    "ST_ANGLE",
    "ST_AREA",
    "ST_AS_BINARY",
    "ST_AS_GEO_JSON",
    "ST_AS_KML",
    "ST_AS_TEXT",
    "ST_AZIMUTH",
    "ST_BOUNDARY",
    "ST_BOUNDING_BOX",
    "ST_BUFFER",
    "ST_BUFFER_WITH_TOLERANCE",
    "ST_CENTROID",
    "ST_CENTROID_AGG",
    "ST_CLOSEST_POINT",
    "ST_CLUSTERDBSCAN",
    "ST_CONTAINS",
    "ST_CONVEXHULL",
    "ST_COVEREDBY",
    "ST_COVERS",
    "ST_DIFFERENCE",
    "ST_DIMENSION",
    "ST_DISJOINT",
    "ST_DISTANCE",
    "ST_DUMP",
    "ST_DUMP_POINTS",
    "ST_DWITHIN",
    "ST_END_POINT",
    "ST_EQUALS",
    "ST_EXTENT",
    "ST_EXTERIORRING",
    "ST_GEOG_FROM_BYTES",
    "ST_GEOG_FROM_GEO_JSON",
    "ST_GEOG_FROM_GEO_JSON_EXT",
    "ST_GEOG_FROM_KML",
    "ST_GEOG_FROM_STRING",
    "ST_GEOG_FROM_TEXT",
    "ST_GEOG_FROM_TEXT_EXT",
    "ST_GEOG_FROM_WKB",
    "ST_GEOG_FROM_WKB_HEX",
    "ST_GEOG_POINT",
    "ST_GEOG_POINT_FROM_GEOHASH",
    "ST_GEOHASH",
    "ST_GEOMETRY_TYPE",
    "ST_INTERIORRINGS",
    "ST_INTERSECTION",
    "ST_INTERSECTS",
    "ST_INTERSECTS_BOX",
    "ST_IS_CLOSED",
    "ST_IS_COLLECTION",
    "ST_IS_EMPTY",
    "ST_IS_RING",
    "ST_LENGTH",
    "ST_LINE_LOCATE_POINT",
    "ST_MAKE_LINE",
    "ST_MAKE_LINE_ARRAY",
    "ST_MAKE_POLYGON",
    "ST_MAKE_POLYGON_ORIENTED",
    "ST_MAX_DISTANCE",
    "ST_NEAREST_NEIGHBORS",
    "ST_NUM_GEOMETRIES",
    "ST_NUM_POINTS",
    "ST_PERIMETER",
    "ST_POINT_N",
    "ST_SIMPLIFY",
    "ST_SNAP_TO_GRID",
    "ST_START_POINT",
    "ST_TOUCHES",
    "ST_UNARY_UNION",
    "ST_UNION",
    "ST_UNION_AGG",
    "ST_UNION_ARRAY",
    "ST_WITHIN",
    "ST_X",
    "ST_Y",
    "SUBSTR_BYTES",
    "SUBSTR_STRING",
    "SUBTRACT_BIGNUMERIC",
    "SUBTRACT_DATE",
    "SUBTRACT_DATETIME",
    "SUBTRACT_DATETIME_INTERVAL",
    "SUBTRACT_DATE_INT",
    "SUBTRACT_DATE_INTERVAL",
    "SUBTRACT_DOUBLE",
    "SUBTRACT_INT",
    "SUBTRACT_INTERVAL_INTERVAL",
    "SUBTRACT_NUMERIC",
    "SUBTRACT_TIME",
    "SUBTRACT_TIMESTAMP",
    "SUBTRACT_TIMESTAMP_INTERVAL",
    "SUBTRACT_UINT",
    "SUM_BIGNUMERIC",
    "SUM_DOUBLE",
    "SUM_INT",
    "SUM_INTERVAL",
    "SUM_NUMERIC",
    "SUM_UINT",
    "TANH_DOUBLE",
    "TAN_DOUBLE",
    "TIMEDIFF_DATE",
    "TIMEDIFF_TIMESTAMP",
    "TIMESTAMP_ADD",
    "TIMESTAMP_BUCKET",
    "TIMESTAMP_DIFF",
    "TIMESTAMP_FROM_DATE",
    "TIMESTAMP_FROM_DATETIME",
    "TIMESTAMP_FROM_INT",
    "TIMESTAMP_FROM_PART_DATE",
    "TIMESTAMP_FROM_PART_DOUBLE",
    "TIMESTAMP_FROM_STRING",
    "TIMESTAMP_FROM_TIMESTAMP",
    "TIMESTAMP_FROM_UNIX_MICROS_INT",
    "TIMESTAMP_FROM_UNIX_MICROS_TIMESTAMP",
    "TIMESTAMP_FROM_UNIX_MILLIS_INT",
    "TIMESTAMP_FROM_UNIX_MILLIS_TIMESTAMP",
    "TIMESTAMP_FROM_UNIX_SECONDS_INT",
    "TIMESTAMP_FROM_UNIX_SECONDS_TIMESTAMP",
    "TIMESTAMP_SUB",
    "TIMESTAMP_TRUNC",
    "TIMESTAMP_TZ_FROM_PART_DATE",
    "TIMESTAMP_TZ_FROM_PART_DOUBLE",
    "TIME_ADD",
    "TIME_DIFF",
    "TIME_FROM_DATETIME",
    "TIME_FROM_HOUR_MINUTE_SECOND",
    "TIME_FROM_PARTS",
    "TIME_FROM_STRING",
    "TIME_FROM_TIME",
    "TIME_FROM_TIMESTAMP",
    "TIME_SLICE_DATE",
    "TIME_SLICE_TIMESTAMP",
    "TIME_SUB",
    "TIME_TRUNC",
    "TO_ARRAY",
    "TO_BASE",
    "TO_BINARY_STRING",
    "TO_BINARY_VARIANT",
    "TO_BOOLEAN_BIGNUMERIC",
    "TO_BOOLEAN_BOOL",
    "TO_BOOLEAN_DOUBLE",
    "TO_BOOLEAN_INT",
    "TO_BOOLEAN_NUMERIC",
    "TO_BOOLEAN_STRING",
    "TO_BOOLEAN_UINT",
    "TO_CODE_POINTS_BYTES",
    "TO_CODE_POINTS_STRING",
    "TO_DATE",
    "TO_DECIMAL_NUMERIC_",
    "TO_DECIMAL_STRING_",
    "TO_DECIMAL_VARIANT_",
    "TO_DOUBLE_BIGNUMERIC",
    "TO_DOUBLE_BOOL",
    "TO_DOUBLE_DOUBLE",
    "TO_DOUBLE_INT",
    "TO_DOUBLE_NUMERIC",
    "TO_DOUBLE_STRING",
    "TO_DOUBLE_UINT",
    "TO_HEX",
    "TO_JSON",
    "TO_JSON_SNOWFLAKE",
    "TO_JSON_STRING",
    "TO_OBJECT_OBJECT",
    "TO_OBJECT_VARIANT",
    "TO_PROTO_BOOL",
    "TO_PROTO_BYTES",
    "TO_PROTO_DATE",
    "TO_PROTO_DOUBLE",
    "TO_PROTO_FLOAT",
    "TO_PROTO_IDEMPOTENT_BOOL",
    "TO_PROTO_IDEMPOTENT_BYTES",
    "TO_PROTO_IDEMPOTENT_DATE",
    "TO_PROTO_IDEMPOTENT_DOUBLE",
    "TO_PROTO_IDEMPOTENT_FLOAT",
    "TO_PROTO_IDEMPOTENT_INT",
    "TO_PROTO_IDEMPOTENT_STRING",
    "TO_PROTO_IDEMPOTENT_TIMESTAMP",
    "TO_PROTO_IDEMPOTENT_TIME_OF_DAY",
    "TO_PROTO_IDEMPOTENT_UINT",
    "TO_PROTO_INT",
    "TO_PROTO_STRING",
    "TO_PROTO_TIME",
    "TO_PROTO_TIMESTAMP",
    "TO_PROTO_UINT",
    "TO_TIMESTAMP_DATE",
    "TO_TIMESTAMP_NUMERIC",
    "TO_TIMESTAMP_STRING",
    "TO_TIMESTAMP_TIMESTAMP",
    "TO_TIMESTAMP_VARIANT",
    "TO_TIME_STRING",
    "TO_TIME_VARIANT",
    "TO_VARCHAR_BIGNUMERIC",
    "TO_VARCHAR_BOOL",
    "TO_VARCHAR_DATE",
    "TO_VARCHAR_DATETIME",
    "TO_VARCHAR_DOUBLE",
    "TO_VARCHAR_FLOAT",
    "TO_VARCHAR_INT",
    "TO_VARCHAR_NUMERIC",
    "TO_VARCHAR_STRING",
    "TO_VARCHAR_TIME",
    "TO_VARCHAR_TIMESTAMP",
    "TO_VARCHAR_UINT",
    "TO_VARCHAR_VARIANT",
    "TO_VARIANT_ARRAY",
    "TO_VARIANT_BIGNUMERIC",
    "TO_VARIANT_BOOL",
    "TO_VARIANT_BYTES",
    "TO_VARIANT_DATE",
    "TO_VARIANT_DATETIME",
    "TO_VARIANT_DOUBLE",
    "TO_VARIANT_FLOAT",
    "TO_VARIANT_INT",
    "TO_VARIANT_NUMERIC",
    "TO_VARIANT_STRING",
    "TO_VARIANT_TIME",
    "TO_VARIANT_TIMESTAMP",
    "TO_VARIANT_UINT",
    "TO_XML_ARRAY",
    "TO_XML_BOOL",
    "TO_XML_FLOAT",
    "TO_XML_INT",
    "TO_XML_OBJECT",
    "TO_XML_VARIANT",
    "TRANSLATE_BYTES",
    "TRANSLATE_STRING",
    "TRIM_BYTES",
    "TRIM_STRING",
    "TRUNC_BIGNUMERIC",
    "TRUNC_DATE",
    "TRUNC_DOUBLE",
    "TRUNC_FLOAT",
    "TRUNC_NUMERIC",
    "TRUNC_WITH_DIGITS_BIGNUMERIC",
    "TRUNC_WITH_DIGITS_DOUBLE",
    "TRUNC_WITH_DIGITS_FLOAT",
    "TRUNC_WITH_DIGITS_NUMERIC",
    "TRY_BASE",
    "TRY_PARSE_JSON_STRING",
    "TRY_PARSE_JSON_VARIANT",
    "TRY_TO_BINARY_STRING",
    "TRY_TO_BOOLEAN_BOOL",
    "TRY_TO_BOOLEAN_STRING",
    "TRY_TO_DATE",
    "TRY_TO_DECIMAL_STRING_",
    "TRY_TO_DOUBLE",
    "TRY_TO_TIME",
    "TRY_TO_TIMESTAMP",
    "TYPEOF",
    "UNARY_MINUS_BIGNUMERIC",
    "UNARY_MINUS_DOUBLE",
    "UNARY_MINUS_FLOAT",
    "UNARY_MINUS_INT",
    "UNARY_MINUS_INTERVAL",
    "UNARY_MINUS_NUMERIC",
    "UNICODE_STRING",
    "UNIX_DATE",
    "UNIX_MICROS_FROM_TIMESTAMP",
    "UNIX_MILLIS_FROM_TIMESTAMP",
    "UNIX_SECONDS_FROM_TIMESTAMP",
    "UPPER_BYTES",
    "UPPER_STRING",
    "VARIANCE_POP",
    "VAR_POP",
    "VAR_POP_BIGNUMERIC",
    "VAR_POP_NUMERIC",
    "VAR_SAMP",
    "VAR_SAMP_BIGNUMERIC",
    "VAR_SAMP_NUMERIC",
    "XMLGET",
    "ZEROIFNULL",
    "ZEROIFNULL_BIGNUMERIC",
    "ZEROIFNULL_DOUBLE",
    "ZEROIFNULL_FLOAT",
    "ZEROIFNULL_INT",
    "ZEROIFNULL_NUMERIC",
    "ZEROIFNULL_UINT"
]



from __future__ import annotations

import re
from typing import Any

from google.cloud import bigquery
from google.cloud.bigquery.dbapi import Cursor as BigQueryDbApiCursor
from google.cloud.bigquery.enums import StandardSqlTypeNames
from google.cloud.bigquery.table import Row
from harlequin import (
    HarlequinAdapter,
    HarlequinConnection,
    HarlequinCursor,
)
from harlequin.autocomplete.completion import HarlequinCompletion
from harlequin.catalog import Catalog, CatalogItem
from harlequin.exception import HarlequinConnectionError, HarlequinQueryError
from textual_fastdatatable.backend import AutoBackendType

from harlequin_bigquery.cli_options import BIGQUERY_ADAPTER_OPTIONS
from harlequin_bigquery.functions import BUILTIN_FUNCTIONS
from harlequin_bigquery.keywords import RESERVED_KEYWORDS

# Abbreviations for column types
COLUMN_TYPE_MAPPING = {
    StandardSqlTypeNames.TYPE_KIND_UNSPECIFIED: "?",
    StandardSqlTypeNames.INT64: "#",
    StandardSqlTypeNames.BOOL: "t/f",
    StandardSqlTypeNames.FLOAT64: "#.#",
    StandardSqlTypeNames.STRING: "s",
    StandardSqlTypeNames.BYTES: "0b",
    StandardSqlTypeNames.TIMESTAMP: "ts",
    StandardSqlTypeNames.DATE: "d",
    StandardSqlTypeNames.TIME: "t",
    StandardSqlTypeNames.DATETIME: "dt",
    StandardSqlTypeNames.INTERVAL: "|-|",
    StandardSqlTypeNames.GEOGRAPHY: "geo",
    StandardSqlTypeNames.NUMERIC: "#.#",
    StandardSqlTypeNames.BIGNUMERIC: "#.#",
    StandardSqlTypeNames.JSON: "{j}",
    StandardSqlTypeNames.ARRAY: "[]",
    StandardSqlTypeNames.STRUCT: "{}",
}


class BigQueryCursor(HarlequinCursor):
    def __init__(self, cursor: BigQueryDbApiCursor, *args: Any, **kwargs: Any) -> None:
        self.cursor = cursor
        self._limit: int | None = None

    def columns(self) -> list[tuple[str, str]]:
        if not self.cursor.query_job:
            raise TypeError("Cursor has no query job")

        result_schema = self.cursor.query_job.result().schema
        fields = []
        for field in result_schema:
            standard_sql_field = field.to_standard_sql()
            if not standard_sql_field.type or not standard_sql_field.type.type_kind:
                type_label = "?"  # Type is unspecified
            else:
                type_label = COLUMN_TYPE_MAPPING[standard_sql_field.type.type_kind]
            fields.append((field.name, type_label))

        return fields

    def set_limit(self, limit: int) -> BigQueryCursor:
        self._limit = limit
        return self

    def fetchall(self) -> AutoBackendType:
        try:
            if not self.cursor.query_job:
                raise RuntimeError("Cursor has no query job")
            if not self.cursor.query_job.destination:
                return []
            if self._limit is None:
                result: list[Row] = self.cursor.fetchall()
            else:
                result = self.cursor.fetchmany(self._limit)
            return [row.values() for row in result]
        except Exception as e:
            raise HarlequinQueryError(
                msg=str(e),
                title="Harlequin encountered an error while executing your query.",
            ) from e


class BigQueryConnection(HarlequinConnection):
    # Abbreviations for table types
    # From the `table_type` field in the INFORMATION_SCHEMA.TABLES table
    # https://cloud.google.com/bigquery/docs/information-schema-tables#schema

    TABLE_TYPE_MAPPING = {
        "BASE TABLE": "t",
        "CLONE": "tc",
        "SNAPSHOT": "ts",
        "VIEW": "v",
        "MATERIALIZED VIEW": "mv",
        "EXTERNAL": "ext",
    }

    def __init__(
        self,
        project: str | None = None,
        location: str | None = None,
        init_message: str = "",
        **_: Any,
    ) -> None:
        self.location = location or "US"
        self.init_message = init_message
        try:
            self.client = bigquery.Client(project=project, location=location)
            self.conn = bigquery.dbapi.Connection(self.client)
        except Exception as e:
            raise HarlequinConnectionError(
                msg=str(e), title="Harlequin could not connect to BigQuery."
            ) from e

    @property
    def project(self) -> str:
        return self.client.project

    def execute(self, query: str) -> BigQueryCursor:
        try:
            cursor = self.conn.cursor()
            cursor.execute(query)
        except Exception as e:
            raise HarlequinQueryError(
                msg=str(e),
                title="Harlequin encountered an error while executing your query.",
            ) from e

        return BigQueryCursor(cursor)

    def get_catalog(self) -> Catalog:
        query = f"""
            select
                datasets.schema_name as dataset_id,
                tables.table_name as table_id,
                tables.table_type as table_type,
                columns,column_name,
                columns.data_type as column_type
            from `{self.project}.region-{self.location}.INFORMATION_SCHEMA.SCHEMATA` datasets
            left join `{self.project}.region-{self.location}.INFORMATION_SCHEMA.TABLES` tables
            on datasets.catalog_name = tables.table_catalog
            and datasets.schema_name = tables.table_schema
            left join `{self.project}.region-{self.location}.INFORMATION_SCHEMA.COLUMNS` columns
            using (table_catalog, table_schema, table_name)
            order by dataset_id, table_id, column_name
        """
        cursor = self.execute(query)
        results = cursor.cursor.fetchall()

        dataset_id = None
        table_id = None
        dataset_items: list[CatalogItem] = []
        table_items: list[CatalogItem] = []
        column_items: list[CatalogItem] = []

        # Iterate in sorted order by dataset, table, then column
        for row in results:
            if row.table_id is None:  # Empty dataset
                pass
            elif table_id is None:
                table_id = row.table_id
            elif row.table_id != table_id:
                table_items.append(
                    CatalogItem(
                        qualified_identifier=f"`{self.project}`.`{dataset_id}`.`{table_id}`",
                        query_name=f"`{table_id}`",
                        label=table_id,
                        type_label=self.TABLE_TYPE_MAPPING[row.table_type],
                        children=column_items,
                    )
                )
                column_items = []
                table_id = row.table_id

            if dataset_id is None:
                dataset_id = row.dataset_id
            elif row.dataset_id != dataset_id:
                dataset_items.append(
                    CatalogItem(
                        qualified_identifier=f"`{self.project}`.`{dataset_id}`",
                        query_name=f"`{dataset_id}`",
                        label=dataset_id,
                        type_label="ds",
                        children=table_items,
                    )
                )
                table_items = []
                dataset_id = row.dataset_id

            if not row.column_name and not row.column_type:
                continue

            # remove anything in <> from the column_type
            column_type_cleaned = re.sub(r"\<.*\>", "", row.column_type)
            column_type_label = COLUMN_TYPE_MAPPING[
                StandardSqlTypeNames(column_type_cleaned)
            ]
            column_items.append(
                CatalogItem(
                    qualified_identifier=f"`{self.project}`.`{row.dataset_id}`.`{row.table_id}`.`{row.column_name}`",
                    query_name=f"`{row.column_name}`",
                    label=row.column_name,
                    type_label=column_type_label,
                )
            )
        else:
            table_items.append(
                CatalogItem(
                    qualified_identifier=f"`{self.project}`.`{dataset_id}`.`{table_id}`",
                    query_name=f"`{table_id}`",
                    label=table_id,
                    type_label=self.TABLE_TYPE_MAPPING[row.table_type],
                    children=column_items,
                )
            )
            dataset_items.append(
                CatalogItem(
                    qualified_identifier=f"`{self.project}`.`{dataset_id}`",
                    query_name=f"`{dataset_id}`",
                    label=dataset_id,
                    type_label="ds",
                    children=table_items,
                )
            )

        return Catalog(items=dataset_items)

    def get_completions(self) -> list[HarlequinCompletion]:
        type_completions = [
            HarlequinCompletion(
                label=str(type_name.value),
                type_label="type",
                value=str(type_name.value),
                priority=1000,
                context=None,
            )
            for type_name in StandardSqlTypeNames
        ]

        keyword_completions = [
            HarlequinCompletion(
                label=keyword,
                type_label="kw",
                value=keyword,
                priority=100,
                context=None,
            )
            for keyword in RESERVED_KEYWORDS
        ]

        # TODO: Get UDFs and routines
        function_completions = [
            HarlequinCompletion(
                label=name,
                type_label="fn",
                value=name,
                priority=1000,
                context=None,
            )
            for name in BUILTIN_FUNCTIONS
        ]

        return [*type_completions, *keyword_completions, *function_completions]


class BigQueryAdapter(HarlequinAdapter):
    ADAPTER_OPTIONS = BIGQUERY_ADAPTER_OPTIONS  # type: ignore

    def __init__(
        self, project: str | None = None, location: str | None = None, **_: Any
    ) -> None:
        self.project = project
        self.location = location

    def connect(self) -> BigQueryConnection:
        conn = BigQueryConnection(project=self.project, location=self.location)
        return conn



from __future__ import annotations

import re

from harlequin.options import TextOption


def is_valid_project(project: str | None) -> tuple[bool, str | None]:
    if project is None:
        return True, None
    is_valid = (
        re.match(r"^[a-z][a-z0-9-]{4,28}[a-z0-9]$", project, flags=re.IGNORECASE)
        is not None
    )
    return (
        is_valid,
        "Must provide a valid project ID" if not is_valid else None,
    )


def is_valid_region(region: str | None) -> tuple[bool, str | None]:
    if region is None:
        return True, None
    is_valid = (
        re.match(r"^[a-z][a-z0-9-]+[a-z0-9]$", region, flags=re.IGNORECASE) is not None
    )
    return (
        is_valid,
        "Must provide a valid region" if not is_valid else None,
    )


project = TextOption(
    name="project",
    description="The project ID to use for the BigQuery connection",
    short_decls=["-p"],
    validator=is_valid_project,
)

location = TextOption(
    name="location",
    description="The location to use for the BigQuery connection",
    short_decls=["-l"],
)

BIGQUERY_ADAPTER_OPTIONS = [project, location]
